<?

 /*****************************************************************************
 * Filename: include/protocol/protocol.php
 *
 * This file is part of the JAS project.
 * Copyright 2003-2015 by Matthijs Schuurman
 *
 * The JAS project is free software under the terms of the
 * GNU Lesser General Public License version 3 or any later version.
 *
 * For information about the license please refer to:
 * LICENSE.txt in the root of the project or http://jas.junerules.com/license/
 *****************************************************************************/


 require_once($_path["include"]."script.php");
 require_once($_path["include"]."socket.php");
 require_once($_path["include"]."errors.php");
 require_once($_path["include"]."data.php");

 declare(ticks=1); //ensure signals are handled immediately

 class Protocol extends Script
 {
  function __construct($server=false)
  {
   $this->connection=new Socket();
   $this->listen=new Socket();

   parent::__construct();
   parent::clear();

   //fork server connections
   $this->fork=true;

   //user/group to run service as
   $this->user=false;
   $this->group=false;

   $this->e->l->read("protocol");

   if($server)
    $this->e->logErrorTypes=$this->e->errorType; //send all errors to stderr

   $this->linebreak="\r\n";

   //update blacklist
   $this->blacklist[]="type";
   $this->blacklist[]="main";
   $this->blacklist[]="setUserGroup";

   //default protocol (tcp, udp or file)
   $this->connectionType($this->connection->connectionTypes[0]);

   //determine protocol type
   if(get_class($this)=="Protocol")
    $this->type=false;
   else if(preg_match("/^Protocol_([A-Za-z]+)$/", get_class($this), $m))
   {
    $this->type=$m[1];
    unset($m);
   }
   else
   {
    $this->e->add("Unable to determine Protocol type from: \"".get_class($this)."\"", "system");
    return;
   }

   //set script variables
   $this->set("split", true); //split read lines by default
   $this->set("linebreak", true); //add linebreak after write by default

   $this->set("connection.default", Array( //default connection properties
    "port" => 23,
    "keepalive" => true, //keep server / client connection alive after main script
    "timeout" => false, //don't timeout listen/read/write
    "reread" => false, //subsequent reads block
   ));

   if($server)
    $this->set("mode", "server", true);
   else
    $this->set("mode", "client", true);

   //default scripts
   if(!$this->parse("connect", "
get mode == server
 listen |url|
  log debug 'New connection from: |connection.shost||:|connection.sport||'
 else
  get shutdown
   log note 'Shutting down'
   exit 0
  else
   log error 'Failed to listen'
   exit false
get mode == client
 connect |url|
  log debug 'Connected to: |url|'
 else
  exit false
exist hello
 execute hello
 else
  exit false
")) return;

   if(!$this->parse("disconnect", "
exist bye
 execute bye
get mode == server
 log debug 'Disconnected from: |connection.shost||:|connection.sport||'
else
 log debug 'Disconnected from: |url|'
disconnect
")) return;

   if(!$this->parse("main", "
server ping
client pong
exit false
")) return;

  }

  function __clone()
  {
   parent::__clone();

   //connected
   if($this->connection->socket||$this->connection->socket6)
    $this->e->add("Connection cloning deactived socket", "warning");

   //listening
   if($this->listen->socket||$this->listen->socket6)
    $this->e->add("Listen cloning deactived socket", "warning");

   //clone objects (and ensure original isn't disconnected)
   $this->connection=clone $this->connection;
   $this->connection->e->clear();
   $this->listen=clone $this->listen;
   $this->listen->e->clear();
  }

  function clear()
  {
   $this->buffer="";
   $this->pipeline="";

   if($this->connection->socket)
    $this->connection->close();
  }

  function connectionType($type)
  {
   if(!in_array($type, $this->connection->connectionTypes))
   {
    $this->e->add($this->type." Protocol unknown connection type: \"".$type."\"", "system");
    return false;
   }

   $this->connectionType=$type;
   return true;
  }

  function parse($name, $script)
  {
   if(!parent::parse($name, $script))
    return false;

   foreach(array_keys($this->commands[$name]) as $step)
   {
    if(
     ($this->get("mode")=="server"&&$this->commands[$name][$step]["type"]=="client")|| //tread client commands output validation
     ($this->get("mode")=="client"&&$this->commands[$name][$step]["type"]=="server") //tread server commands output validation
    )
    {
     $this->commands[$name][$step]["output"]=$this->commands[$name][$step]["input"];
     $this->commands[$name][$step]["input"]=false;

     //remove ' from output validation
     if(is_array($this->commands[$name][$step]["output"]))
      $this->commands[$name][$step]["output"]["value"]=str_replace("'", "", $this->commands[$name][$step]["output"]["value"]);
     else
      $this->commands[$name][$step]["output"]=str_replace("'", "", $this->commands[$name][$step]["output"]);
    }
   }

   return true;
  }

  function main()
  {
   if($this->get("mode")=="server")
   {
    if(!$this->execute("connect")) //listen for new connections
     return false;

    while($this->execute("main")) //keep running main loop till failure
     if(!$this->get("connection.keepalive")) //or till keepalive is disabled
      break;

    if(!$this->execute("disconnect"))
     return false;
   }
   else //client
   {
    if(!$this->connection->socket) //not yet connected
     if(!$this->execute("connect")) //connect to server
      return false;

    if(!$this->execute("main")) //run main script
    {
     if($this->get("connection.keepalive")) //connection may have timed out
     {
      //reconnect
      if(!$this->execute("disconnect")||!$this->execute("connect"))
       return false;

      if(!$this->execute("main")) //rerun main script
       return false; //other failure
     }
     else //no keep alive
      return false; //other failure
    }

    if(!$this->get("connection.keepalive"))
     if(!$this->execute("disconnect")) //disconnect
      return false;
    //else don't disconnect: subsequent main calls will reuse existing connections (destroying $this will disconnect)
   }

   return true;
  }

  function setUserGroup($user=false, $group=false)
  {
   if($this->get("mode")!="server") //for server only
   {
    $this->e->add($this->type." Protocol unable to set user/group in client mode", "error");
    return false;
   }

   $this->user=$user;
   $this->group=$group;

   return true;
  }

  function setFork($fork)
  {
   if($this->get("mode")!="server") //for server only
   {
    $this->e->add($this->type." Protocol unable to set fork in client mode", "error");
    return false;
   }

   $this->fork=($fork?true:false);
   return true;
  }


  /*
   * Script commands
   */

  function listen($url=false)
  {
   if($this->get("mode")!="server") //for server only
    return false;

   if(!$this->listen->socket&&!$this->listen->socket6) //already listening?
   {
    if($this->get("connection.default.keepalive")&&!$this->fork)
     $this->e->add($this->type." Protocol ".$this->get("mode")." keep alive isn't advised in non-fork mode", "warning");

    $this->set("shutdown", false);

    if(!$url)
     $url=strtolower($this->type)."://localhost:".$this->get("connection.default.port");

    if(!($listen=$this->listen->URL($url)))
    {
     $this->e->add($this->listen->e);
     return false;
    }

    if(!$listen["protocol"])
     $listen["protocol"]=strtolower($this->type);

    if(!$listen["port"])
     $listen["port"]=$this->get("connection.default.port");

    $listen["reread"]=$this->get("connection.default.reread");

    if(strtolower($listen["protocol"])!=strtolower($this->type))
    {
     $this->e->add($this->type." Protocol ".$this->get("mode")." unsupported protocol: \"".$listen["protocol"]."\"", "error");
     return false;
    }

    if(preg_match("/^[0-F:]+$/i", $listen["host"])) //ipv6 address
     $url=$this->connectionType."://[".$listen["host"]."]:".$listen["port"];
    else //hostname / ipv4 address
     $url=$this->connectionType."://".$listen["host"].":".$listen["port"];

    if(!$this->listen->setReRead($listen["reread"]))
    {
     $this->e->add($this->type." Protocol unable to set reread for: \"".$listen["url"]."\"", "error");
     $this->e->add($this->listen->e);
     return false;
    }

    if(!$this->listen->listen($url, $this->user, $this->group))
    {
     $this->e->add($this->type." Protocol unable to listen to: \"".$listen["url"]."\"", "error");
     $this->e->add($this->listen->e);
     return false;
    }

    //set listen connection properties
    $this->set("listen", $listen);
   }

   $this->buffer="";
   $this->pipeline="";

   if(($s=$this->listen->accept($this->fork))===false)
   {
    if(isset($this->listen->acceptShutdown))
     $this->set("shutdown", $this->listen->acceptShutdown); //server shutdown

    $this->e->add($this->listen->e);
    return false;
   }

   //set new data socket
   unset($this->connection);
   $this->connection=&$s;

   //set connection properties
   $this->set("connection", array_merge(Array("default" => $this->get("connection.default")), $this->get("connection.default"), $this->get("listen")));

   $this->set("connection.host", $this->connection->host);
   $this->set("connection.port", $this->connection->port);
   $this->set("connection.shost", $this->connection->shost);
   $this->set("connection.sport", $this->connection->sport);

   $this->set("connection.time", round(microtime(true)-$this->connection->startTime, 5));
   $this->set("connection.status", "connected");


   //sync timeouts
   if($this->get("connection.timeout")!==$this->connection->getTimeout())
   {
    if(!$this->connection->setTimeout($this->get("connection.timeout")))
     $this->e->add($this->connection->e);

    $this->set("connection.timeout", $this->connection->getTimeout());
   }

   $this->set("connection.reread", $this->connection->getReRead());

   return true;
  }

  function connect($url)
  {
   if($this->get("mode")!="client") //for client only
    return false;

   if($this->connection->socket)
   {
    $this->e->add($this->type." Protocol ".$this->get("mode")." failed, already connected to: \"".$this->get("connection.url")."\"", "error");
    return false;
   }

   $this->buffer="";
   $this->pipeline="";

   if(!($connection=$this->connection->URL($url)))
   {
    $this->e->add($this->connection->e);
    return false;
   }

   if(!$connection["protocol"])
    $connection["protocol"]=strtolower($this->type);

   if(!$connection["port"])
    $connection["port"]=$this->get("connection.default.port");

   $connection["reread"]=$this->get("connection.default.reread");

   if(strtolower($connection["protocol"])!=strtolower($this->type))
   {
    $this->e->add($this->type." Protocol ".$this->get("mode")." unsupported protocol: \"".$connection["protocol"]."\"", "error");
    return false;
   }

   if(preg_match("/^[0-F:]+$/i", $connection["host"])) //ipv6 address
    $url=$this->connectionType."://[".$connection["host"]."]:".$connection["port"];
   else //hostname / ipv4 address
    $url=$this->connectionType."://".$connection["host"].":".$connection["port"];

   if(!$this->connection->setReRead($connection["reread"]))
   {
    $this->e->add($this->type." Protocol unable to set reread for: \"".$connection["url"]."\"", "error");
    $this->e->add($this->listen->e);
    return false;
   }

   if(!$this->connection->connect($url))
   {
    $this->e->add($this->type." Protocol unable to connect to: \"".$connection["url"]."\"", "error");
    $this->e->add($this->connection->e);
    return false;
   }

   //set connection properties
   $this->set("connection", array_merge(Array("default" => $this->get("connection.default")), $this->get("connection.default"), $connection));

   $this->set("connection.shost", $this->connection->shost);
   $this->set("connection.sport", $this->connection->sport);

   $this->set("connection.time", round(microtime(true)-$this->connection->startTime, 5));
   $this->set("connection.status", "connected");

   //sync timeouts
   if($this->get("connection.timeout")!==$this->connection->getTimeout())
   {
    if(!$this->connection->setTimeout($this->get("connection.timeout")))
     $this->e->add($this->connection->e);

    $this->set("connection.timeout", $this->connection->getTimeout());
   }

   $this->set("connection.reread", $this->connection->getReRead());

   return true;
  }

  function disconnect()
  {
   if($this->connection->socket)
   {
    if(!$this->connection->close())
     return false;

    $this->set("connection.time", round(microtime(true)-$this->connection->startTime, 5));
    $this->set("connection.status", "disconnected");
   }

   return true;
  }

  function server($input=false)
  {
   if($this->get("mode")=="server")
    return $this->write($input);
   else //client
    return $this->read();
  }

  function client($input=false)
  {
   if($this->get("mode")=="server")
    return $this->read();
   else //client
    return $this->write($input);
  }

  function read()
  {
   if(!$this->connection->socket)
   {
    $this->e->add($this->type." Protocol ".$this->get("mode")." read failed, not connected", "error");
    return false;
   }

   //sync timeouts
   if($this->get("connection.timeout")!==$this->connection->getTimeout())
   {
    if(!$this->connection->setTimeout($this->get("connection.timeout")))
     $this->e->add($this->connection->e);

    $this->set("connection.timeout", $this->connection->getTimeout());
   }

   //sync reread
   if($this->get("connection.reread")!==$this->connection->getReRead())
   {
    if(!$this->connection->setReRead($this->get("connection.reread")))
     $this->e->add($this->connection->e);

    $this->set("connection.reread", $this->connection->getReRead());
   }

   //receive data
   if(($data=$this->connection->read())===false)
   {
    $this->e->add($this->type." Protocol ".$this->get("mode")." read failed for: \"".$this->get("connection.url")."\"", "error");
    return false;
   }

   //set connection status and time
   $this->set("connection.time", round(microtime(true)-$this->connection->startTime, 5));
   if($this->get("connection.status")!="active")
    $this->set("connection.status", "active");

   if($data===""&&$this->connection->getTimeout()!==false&&$this->connection->getReRead()!==false) //no data + timeout set
    return false; //timeout

   //append data to buffer (keep for reuse later)
   $this->buffer.=$data;

   if(!$this->get("split")) //not in split mode
    return $this->buffer; //return raw buffer

   //split mode
   $data=rtrim($this->buffer, $this->linebreak);

   //convert multiline to array
   if(strpos($data, $this->linebreak)!==false)
    $data=explode($this->linebreak, $data);

   return $data;
  }

  function write($data)
  {
   if(!$this->connection->socket)
   {
    $this->e->add($this->type." Protocol ".$this->get("mode")." write failed, not connected", "error", $data);
    return false;
   }

   //sync timeouts
   if($this->get("connection.timeout")!==$this->connection->getTimeout())
   {
    if(!$this->connection->setTimeout($this->get("connection.timeout")))
     $this->e->add($this->connection->e);

    $this->set("connection.timeout", $this->connection->getTimeout());
   }

   $this->buffer="";

   //convert array to multiline
   if(is_array($data))
    $data=implode($this->linebreak, $data);

   if($this->get("linebreak"))
    $data.=$this->linebreak;

   //send data
   if(!$this->connection->write($data))
   {
    $this->e->add($this->type." Protocol ".$this->get("mode")." write failed for: \"".$this->get("connection.url")."\"", "error");
    return false;
   }

   //set connection status and time
   $this->set("connection.time", round(microtime(true)-$this->connection->startTime, 5));
   if($this->get("connection.status")!="active")
    $this->set("connection.status", "active");

   return true;
  }

 }

?>
