<?

 /*****************************************************************************
 * Filename: include/socket.php
 *
 * This file is part of the JAS project.
 * Copyright 2003-2015 by Matthijs Schuurman
 *
 * The JAS project is free software under the terms of the
 * GNU Lesser General Public License version 3 or any later version.
 *
 * For information about the license please refer to:
 * LICENSE.txt in the root of the project or http://jas.junerules.com/license/
 *****************************************************************************/


 require_once($_path["include"]."errors.php");
 require_once($_path["include"]."mp.php");

 declare(ticks=1); //ensure signals are handled immediately

 class Socket
 {
  function __construct()
  {
   $this->e=new Errors();
   $this->e->l->read("socket");

   $this->ipv4=true;
   $this->ipv6=true;

   $this->blocking=true;
   $this->setTimeout(); //no timeout (i.e. blocking)
   $this->backlog=10; //max connections in backlog

   $this->reread=Array( //reread ensures subsequent reads won't block
    "enabled" => false,
    "timeout" => 0.0001,
    "counter" => 0,
   );

   $this->readBytes=1024; //number of bytes to read per read call
   $this->readType=PHP_BINARY_READ; //read type: PHP_BINARY_READ or PHP_NORMAL_READ

   $this->connectionTypes=Array("tcp", "udp", "file");

   $this->socket=false;
   $this->socket6=false;

   $this->clear();
  }

  function __clone()
  {
   //clone objects
   $this->e=clone $this->e;

   //connected/listening
   if($this->socket||$this->socket6)
   {
    $this->e->add("Socket cloning deactived socket", "warning");

    //reset object (and ensure original isn't disconnected)
    $this->socket=false;
    $this->socket6=false;

    $this->clear();
   }
  }

  function clear()
  {
   if($this->socket||$this->socket6)
    $this->close();

   $this->mode=false;

   $this->file=false;
   $this->host=false;
   $this->port=0;
   $this->shost=false;
   $this->sport=0;

   $this->bytesSend=0;
   $this->bytesReceived=0;

   $this->connectionTime=0;
   $this->startTime=0;

   $this->errno=false;
   $this->error="";
  }

  function error()
  {
   if($this->socket) //socket initialized
   {
    if(!($this->errno=@socket_last_error($this->socket))) //get socket specific error
     $this->errno=@socket_last_error(); //get non socket specific error
   }
   else if($this->socket6) //socket initialized
   {
    if(!($this->errno=@socket_last_error($this->socket6))) //get socket specific error
     $this->errno=@socket_last_error(); //get non socket specific error
   }
   else //no socket
    $this->errno=@socket_last_error(); //get non socket specific error

   socket_clear_error(); //clear socket errors

   if($this->errno)
    $this->error=@socket_strerror($this->errno);
   else
    $this->error="";

   return $this->errno;
  }

  function inet($ipv4, $ipv6)
  {
   if($this->socket||$this->socket6)
   {
    $this->e->add("Socket already active, can't set inet mode", "error");
    return false;
   }

   if(!$ipv4&&!$ipv6)
   {
    $this->e->add("At least IPv4 or IPv6 should be allowed", "error");
    return false;
   }

   if($ipv4)
    $this->ipv4=true;
   else
    $this->ipv4=false;

   if($ipv6)
    $this->ipv6=true;
   else
    $this->ipv6=false;

   return true;
  }


  function setBlocking($blocking)
  {
   if($blocking)
    $this->blocking=true;
   else
    $this->blocking=false;

   if($this->socket||$this->socket6)
   {
    if($this->blocking)
    {
     if(
      ($this->socket&&!@socket_set_block($this->socket))||
      ($this->socket6&&!@socket_set_block($this->socket6))
     )
     {
      $this->error();
      $this->e->add("Unable to set connection to blocking", "error", "[".$this->errno."] ".$this->error);
      return false;
     }
    }
    else //non blocking
    {
     if(
      ($this->socket&&!@socket_set_nonblock($this->socket))||
      ($this->socket6&&!@socket_set_nonblock($this->socket6))
     )
     {
      $this->error();
      $this->e->add("Unable to set connection to non-blocking", "error", "[".$this->errno."] ".$this->error);
      return false;
     }
    }
   }

   return true;
  }


  function setTimeout($timeout=false)
  {
   if(!is_float($timeout)&&!is_int($timeout)&&$timeout!==false)
   {
    $this->e->add("Timeout should be a float or false", "error");
    return false;
   }

   if($timeout===false) //no timeout
    $this->timeout=Array(NULL, NULL); //block
   else //float value
   {
    $this->timeout=Array();
    $this->timeout[0]=(int)floor($timeout); //seconds
    $this->timeout[1]=(int)(($timeout-$this->timeout[0])*1000000);
   }

   return true;
  }

  function getTimeout()
  {
   if($this->timeout[0]!==NULL) //float value
    return (float)$this->timeout[0]+($this->timeout[1]/1000000);

   return false;
  }


  function setReRead($enabled)
  {
   $this->reread["enabled"]=($enabled?true:false);
   return true;
  }

  function getReRead()
  {
   return $this->reread["enabled"];
  }

  function resetReRead()
  {
   $this->reread["counter"]=0;
   return true;
  }


  function setReadBytes($bytes)
  {
   if(!is_int($bytes))
   {
    $this->e->add("Read bytes should be an integer", "error");
    return false;
   }

   $this->readBytes=$bytes;
   return true;
  }

  function setReadType($type)
  {
   if($type=="binary"||$type==PHP_BINARY_READ)
   {
    $this->readType=PHP_BINARY_READ;
    return true;
   }
   else if($type=="normal"||$type==PHP_NORMAL_READ) //stops at \n or \r
   {
    $this->readType=PHP_NORMAL_READ;
    return true;
   }

   $this->e->add("Unknown read type: \"".$type."\"", "error");
   return false;
  }


  function setOption($key, $value)
  {
   if(!$this->socket&&!$this->socket6)
   {
    $this->e->add("Socket not active, refusing to set option", "error");
    return false;
   }

   if(
    ($this->socket&&!@socket_set_option($this->socket, SOCK_STREAM, $key, $value))||
    ($this->socket6&&!@socket_set_option($this->socket6, SOCK_STREAM, $key, $value))
   )
   {
    $this->error();
    $this->e->add("Unable to set socket option", "error", "[".$this->errno."] ".$this->error);
    return false;
   }

   return true;
  }

  function getOption($key)
  {
   if(!$this->socket&&!$this->socket6)
   {
    $this->e->add("Socket not active, refusing to get option", "error");
    return false;
   }

   if(
    ($this->socket&&!($value=@socket_get_option($this->socket, SOCK_STREAM, $key)))||
    ($this->socket6&&!($value=@socket_get_option($this->socket6, SOCK_STREAM, $key)))
   )
   {
    $this->error();
    $this->e->add("Unable to get socket option", "error", "[".$this->errno."] ".$this->error);
    return false;
   }

   return $value;
  }


  function URL($url) //file:///dir/file.sock, udp://192.168.0.1:234, tcp://[::1]:443 or www.test.com:80 (tcp is default)
  {
   if(!preg_match("!^(([a-z]+)://)?(([^@]+)@)?([a-z0-9\.]+|\[[0-F:]+\])?(:(\d+))?(/[^\?]*)?(\?.+)?$!i", $url, $m))
   {
    $this->e->add("Socket invalid URL: \"".$url."\"", "error");
    return false;
   }

   $url=Array("url" => $url);

   if($m[2])
    $url["protocol"]=$m[2];
   else
    $url["protocol"]=false;

   if($m[4])
   {
    $tmp=explode(":", $m[4]);
    $url["user"]=$tmp[0];
    $url["pass"]=(isset($tmp[1])?$tmp[1]:false);
   }
   else
   {
    $url["user"]=false;
    $url["pass"]=false;
   }

   if($m[5])
   {
    if(preg_match("/^\[([0-F:]+)\]$/i", $m[5], $m2)) //ipv6 address
     $url["host"]=$m2[1];
    else //hostname / ipv4 address
     $url["host"]=$m[5];
   }
   else
    $url["host"]=false;


   if(isset($m[7])&&$m[7])
    $url["port"]=$m[7];
   else
    $url["port"]=false;

   if(isset($m[8])&&$m[8])
    $url["path"]=$m[8];
   else
    $url["path"]=false;

   if(isset($m[9])&&$m[9])
   {
    $url["values"]=Array();
    $tmp=substr($m[9], 1); //remove ?
    foreach(explode("&", $tmp) as $kv) //get each key/value pair
    {
     $kv=explode("=", $kv);
     if(isset($kv[1]))
      $url["values"][]=Array("key" => $kv[0], "value" => $kv[1]);
     else
      $url["values"][]=Array("key" => $kv[0], "value" => "");
    }
   }
   else
    $url["values"]=false;

   if($url["protocol"]=="file"&&$url["host"]) //relative path for file
   {
    $url["path"]=$url["host"].$url["path"]; //append host to path
    $url["host"]=false; //no host
   }

   return $url;
  }

  function connect($url)
  {
   if($this->socket||$this->socket6)
   {
    $this->e->add("Socket already active, refusing to connect", "error");
    return false;
   }

   if($this->reread["enabled"])
    $this->resetReRead();

   if(!($url=$this->URL($url)))
    return false;

   if(!$url["protocol"])
    $url["protocol"]=$this->connectionTypes[0];

   if(!in_array($url["protocol"], $this->connectionTypes))
   {
    $this->e->add("Unknown connection type: \"".$url["protocol"]."\"", "error");
    return false;
   }

   $this->clear();
   $this->mode="connection";
   $this->startTime=microtime(true);

   if($url["protocol"]=="file")
   {
    if(!file_exists($url["path"]))
    {
     $this->e->add("Connection failed: socket: \"".$url["path"]."\" doesn't exists", "error");
     return false;
    }

    if(filetype($url["path"])!="socket")
    {
     $this->e->add("Connection failed: file: \"".$url["path"]."\" isn't a socket", "error");
     return false;
    }

    $this->file=$url["path"];
    $this->host=false;
    $this->port=0;

    if(($this->socket=@socket_create(AF_UNIX, SOCK_STREAM, 0))===false)
    {
     $this->error();
     $this->e->add("Connection failed: unable to create ".$url["protocol"]." socket", "error", "[".$this->errno."] ".$this->error);
     return false;
    }

    if(!@socket_connect($this->socket, $this->file))
    {
     $this->error();
     $this->e->add("Connection failed: unable to connect to: \"".$this->file."\"", "error", "[".$this->errno."] ".$this->error);
     $this->close();
     return false;
    }
   }
   else //tcp/udp
   {
    if(preg_match("/^\d+\.\d+\.\d+\.\d+$/", $url["host"])) //ipv4
     $inet="ipv4";
    else if(preg_match("/^[0-F\:]+$/i", $url["host"])) //ipv6
     $inet="ipv6";
    else //host name
     $inet="both";

    if($inet=="ipv4"&&!$this->ipv4)
    {
     $this->e->add("Listen failed: IPv4 is disabled", "error");
     return false;
    }

    if($inet=="ipv6"&&!$this->ipv6)
    {
     $this->e->add("Listen failed: IPv6 is disabled", "error");
     return false;
    }

    if(!$url["port"])
    {
     $this->e->add("Connection failed: port number not set", "error");
     return false;
    }

    $this->file=false;
    $this->host=$url["host"];
    $this->port=$url["port"];

    if($url["protocol"]=="udp")
     $protocol=SOL_UDP;
    else
     $protocol=SOL_TCP;

    if($this->ipv6&&($inet=="ipv6"||$inet=="both"))
    {
     if($this->host=="localhost") //localhost ipv6 hack (may not have an address associated)
      $host="::1";
     else
      $host=$this->host;

     if(($this->socket=@socket_create(AF_INET6, SOCK_STREAM, $protocol))===false)
     {
      if(!$this->ipv4||$inet=="ipv6") //no retry
      {
       $this->error();
       $this->e->add("Connection failed: unable to create ".$url["protocol"]." socket", "error", "[".$this->errno."] ".$this->error);
       return false;
      }
     }
     else if(!@socket_connect($this->socket, $host, $this->port))
     {
      if(!$this->ipv4||$inet=="ipv6") //no retry
      {
       $this->error();
       $this->e->add("Connection failed: unable to connect to: \"".($this->host?$this->host.":".$this->port:$this->file)."\"", "error", "[".$this->errno."] ".$this->error);
       $this->close();
       return false;
      }

      $this->close();
     }
    }

    if($this->ipv4&&($inet=="ipv4"||$inet=="both"))
    {
     if(!$this->socket) //IPv4 retry or first try if ipv6 is disabled
     {
      if(($this->socket=@socket_create(AF_INET, SOCK_STREAM, $protocol))===false)
      {
       $this->error();
       $this->e->add("Connection failed: unable to create ".$url["protocol"]." socket", "error", "[".$this->errno."] ".$this->error);
       return false;
      }

      if(!@socket_connect($this->socket, $this->host, $this->port))
      {
       $this->error();
       $this->e->add("Connection failed: unable to connect to: \"".($this->host?$this->host.":".$this->port:$this->file)."\"", "error", "[".$this->errno."] ".$this->error);
       $this->close();
       return false;
      }
     }
    }

    //get source info for tcp/udp socket
    socket_getsockname($this->socket, $this->shost, $this->sport);
   }

   if(!$this->setBlocking($this->blocking))
   {
    $this->close();
    return false;
   }

   return true;
  }


  function listen($url, $user=false, $group=false)
  {
   if($this->socket||$this->socket6)
   {
    $this->e->add("Socket already active, refusing to listen", "error");
    return false;
   }

   if($this->reread["enabled"])
    $this->resetReRead();

   if(!($url=$this->URL($url)))
    return false;

   if(!$url["protocol"])
    $url["protocol"]=$this->connectionTypes[0];

   if(!in_array($url["protocol"], $this->connectionTypes))
   {
    $this->e->add("Unknown connection type: \"".$url["protocol"]."\"", "error");
    return false;
   }

   $this->clear();
   $this->mode="listen";
   $this->startTime=microtime(true);

   if($url["protocol"]=="file")
   {
    if(file_exists($url["path"])) //done before setting filename, otherwise close will unlink the file
    {
     $this->e->add("Listen failed: socket: \"".$url["path"]."\" already exists", "error");
     return false;
    }

    $this->file=$url["path"];
    $this->host=false;
    $this->port=0;

    if(($this->socket=@socket_create(AF_UNIX, SOCK_STREAM, 0))===false)
    {
     $this->error();
     $this->e->add("Listen failed: unable to create ".$url["protocol"]." socket", "error", "[".$this->errno."] ".$this->error);
     return false;
    }

    if(!@socket_bind($this->socket, $this->file))
    {
     $this->error();
     $this->e->add("Listen failed: unable to bind on: \"".$this->file."\"", "error", "[".$this->errno."] ".$this->error);
     $this->close();
     return false;
    }

    if(!FS::setModeS($this->file, "rw", "", ""))
     $this->e->add("Listen: unable to change file mode on: \"".$this->file."\"", "error");
   }
   else //tcp/udp
   {
    if(!$url["host"]) //listen on all
     $url["host"]="::";

    if(preg_match("/^\d+\.\d+\.\d+\.\d+$/", $url["host"])) //ipv4
     $inet="ipv4";
    else if(preg_match("/^[0-F\:]+$/i", $url["host"])) //ipv6
     $inet="ipv6";
    else //host name
     $inet="both";

    if($inet=="ipv4"&&!$this->ipv4)
    {
     $this->e->add("Listen failed: IPv4 is disabled", "error");
     return false;
    }

    if($inet=="ipv6"&&!$this->ipv6)
    {
     $this->e->add("Listen failed: IPv6 is disabled", "error");
     return false;
    }

    if(!$url["port"])
    {
     $this->e->add("Listen failed: port number not set", "error");
     return false;
    }

    $this->file=false;
    $this->host=$url["host"];
    $this->port=$url["port"];

    if($url["protocol"]=="udp")
     $protocol=SOL_UDP;
    else
     $protocol=SOL_TCP;

    if($this->ipv6&&($inet=="ipv6"||$inet=="both"))
    {
     if($this->host=="localhost") //localhost ipv6 hack (may not have an address associated)
      $host="::1";
     else
      $host=$this->host;

     if(($this->socket6=@socket_create(AF_INET6, SOCK_STREAM, $protocol))===false)
     {
      if(!$this->ipv4||$inet=="ipv6") //only ipv6
      {
       $this->error();
       $this->e->add("Listen failed: unable to create ".$url["protocol"]." socket", "error", "[".$this->errno."] ".$this->error);
       return false;
      }
     }
     else if(!@socket_set_option($this->socket6, SOCK_STREAM, SO_REUSEADDR, 1))
     {
      if(!$this->ipv4||$inet=="ipv6") //only ipv6
      {
       $this->error();
       $this->e->add("Unable to set socket reuse option", "error", "[".$this->errno."] ".$this->error);
       $this->close();
       return false;
      }

      @socket_close($this->socket6); //close socket
      $this->socket6=false;
     }
     else if(!@socket_bind($this->socket6, $host, $this->port))
     {
      if(!$this->ipv4||$inet=="ipv6") //only ipv6
      {
       $this->error();
       $this->e->add("Listen failed: unable to bind on: \"".($this->host?$this->host.":".$this->port:$this->file)."\"", "error", "[".$this->errno."] ".$this->error);
       $this->close();
       return false;
      }

      @socket_close($this->socket6); //close socket
      $this->socket6=false;
     }
    }

    if($this->ipv4&&($inet=="ipv4"||$inet=="both"))
    {
     if(($this->socket=@socket_create(AF_INET, SOCK_STREAM, $protocol))===false)
     {
      if(!$this->socket6) //no ipv6
      {
       $this->error();
       $this->e->add("Listen failed: unable to create ".$url["protocol"]." socket", "error", "[".$this->errno."] ".$this->error);
       return false;
      }
     }
     else if(!@socket_set_option($this->socket, SOCK_STREAM, SO_REUSEADDR, 1))
     {
      if(!$this->socket6) //no ipv6
      {
       $this->error();
       $this->e->add("Unable to set socket reuse option", "error", "[".$this->errno."] ".$this->error);
       $this->close();
       return false;
      }

      @socket_close($this->socket); //close socket
      $this->socket=false;
     }
     else if(!@socket_bind($this->socket, $this->host, $this->port))
     {
      if(!$this->socket6) //no ipv6
      {
       $this->error();
       $this->e->add("Listen failed: unable to bind on: \"".($this->host?$this->host.":".$this->port:$this->file)."\"", "error", "[".$this->errno."] ".$this->error);
       $this->close();
       return false;
      }

      @socket_close($this->socket); //close socket
      $this->socket=false;
     }
    }
   }

   //drop priveledges
   if($group&&!preg_match("/^\d+$/", $group)) //group name
    $gid=posix_getgrnam($group)["gid"]; //get id
   else
    $gid=$group;

   if($user&&!preg_match("/^\d+$/", $user)) //user name
    $uid=posix_getpwnam($user)["uid"]; //get id
   else
    $uid=$user;

   if($gid&&posix_getgid()!=$gid)
   {
    if(!posix_setgid($gid))
    {
     $this->e->add("Listen can't become group: \"".$group."\"", "error");
     $this->close();
     return false;
    }
   }

   if($uid&&posix_getuid()!=$uid)
   {
    if(!posix_setuid($uid))
    {
     $this->e->add("Listen can't become user: \"".$user."\"", "error");
     $this->close();
     return false;
    }
   }

   //listen
   if(
    ($this->socket&&!@socket_listen($this->socket, $this->backlog))||
    ($this->socket6&&!@socket_listen($this->socket6, $this->backlog))
   )
   {
    $this->error();
    $this->e->add("Listen failed: unable to listen on: \"".($this->host?$this->host.":".$this->port:$this->file)."\"", "error", "[".$this->errno."] ".$this->error);
    $this->close();
    return false;
   }

   if(!$this->setBlocking($this->blocking))
   {
    $this->close();
    return false;
   }

   return true;
  }

  function accept($fork=false)
  {
   if(!$this->socket&&!$this->socket6)
   {
    $this->e->add("Socket not active, refusing to accept connections", "error");
    return false;
   }

   if($this->mode!="listen")
   {
    $this->e->add("Socket not in listen mode, refusing to accept connections", "error");
    return false;
   }

   //prepare select
   $write=NULL;
   $exception=NULL;

   if($fork)
   {
    $mp=new MP();

    if( //shutdown signals
     !$mp->setSignal("interrupt", $this, "acceptShutdown")||
     !$mp->setSignal("terminate", $this, "acceptShutdown")
    )
    {
     $this->e->add($mp->e);
     return false;
    }

    $this->acceptShutdown=false;
    while(!$this->acceptShutdown)
    {
     //prepare select
     $read=Array();
     if($this->socket)
      $read[]=$this->socket;
     if($this->socket6)
      $read[]=$this->socket6;

     if(($sockets=@socket_select($read, $write, $exception, $this->timeout[0], $this->timeout[1]))===false)
     {
      if($this->error()==4) //interrupt
       continue; //continue accepting clients

      $this->e->add("Connection accept failed: unable to select on socket", "error", "[".$this->errno."] ".$this->error);
      return false;
     }

     if($sockets===0) //Timeout/non blocking: no new clients
      continue; //continue accepting clients

     foreach($read as $sock)
     {
      if(!($socket=@socket_accept($sock)))
      {
       if($this->error()) //signal interrupt can trigger a failure, errno is 0 returned
        $this->e->add("Unable to accept", "error", "[".$this->errno."] ".$this->error);

       return false;
      }

      if(!($process=$mp->fork()))
      {
       $this->e->add("Unable to fork client", "error");
       return false;
      }

      if($process=="child") //child handles connection
       break;
     }

     if($process=="child") //child handles connection
     {
      //ensure child doesn't shutdown listen socket
      $this->socket=false;
      $this->socket6=false;

      break;
     }
     //else parent, keep accepting
    }

    if($this->acceptShutdown) //shutting down
    {
     $mp->sendSignal("terminate", "childs"); //tell kids it's bedtime
     $mp->wait("childs"); //wait for kids to go to bed
     return false;
    }
   }
   else //don't fork
   {
    //prepare select
    $read=Array();
    if($this->socket)
     $read[]=$this->socket;
    if($this->socket6)
     $read[]=$this->socket6;

    if(($sockets=@socket_select($read, $write, $exception, $this->timeout[0], $this->timeout[1]))===false)
    {
     if($this->error()) //signal interrupt can trigger a failure, errno is 0 returned
      $this->e->add("Connection accept failed: unable to select on socket", "error", "[".$this->errno."] ".$this->error);

     return false;
    }

    if($sockets===0) //Timeout/non blocking: no new clients
     return false; //done accepting clients

    foreach($read as $sock)
    {
     if(!($socket=@socket_accept($sock)))
     {
      if($this->error()) //signal interrupt can trigger a failure, errno is 0 returned
       $this->e->add("Unable to accept", "error", "[".$this->errno."] ".$this->error);

      return false;
     }

     break; //only accept new connection on first socket, others handled later
    }
   }

   $s=new Socket();
   $s->socket=$socket;
   $s->socket6=false;
   $s->mode="connection";

   $s->startTime=microtime(true);
   $s->connectionTime=0;

   $s->ipv4=$this->ipv4;
   $s->ipv6=$this->ipv6;

   $s->host=$this->host;
   $s->port=$this->port;
   $s->file=$this->file;

   if($s->host) //get source info for tcp/udp socket
   {
    socket_getsockname($s->socket, $s->host, $s->port);
    socket_getpeername($s->socket, $s->shost, $s->sport);

    //proper ipv4 address when listening on ::
    if(preg_match("/^::ffff:(\d+\.\d+\.\d+\.\d+)$/i", $s->host, $m))
     $s->host=$m[1];
    if(preg_match("/^::ffff:(\d+\.\d+\.\d+\.\d+)$/i", $s->shost, $m))
     $s->shost=$m[1];
   }

   $s->setBlocking($this->blocking);
   $s->setTimeout($this->getTimeout());
   $s->setReRead($this->getReRead());

   $s->readBytes=$this->readBytes;
   $s->readType=$this->readType;

   $s->bytesReceived=0;
   $s->bytesSend=0;

   return $s;
  }

  function acceptShutdown()
  {
   $this->acceptShutdown=true;
  }


  function read()
  {
   if(!$this->socket)
   {
    $this->e->add("Socket not active, refusing to read", "error");
    return false;
   }

   $read=Array($this->socket);
   $write=NULL;
   $exception=NULL;

   if($this->reread["enabled"]&&$this->reread["counter"])
   {
    $timeout=$this->getTimeout(); //store original timeout
    $this->setTimeout($this->reread["timeout"]); //set reread timeout
   }

   if(($sockets=@socket_select($read, $write, $exception, $this->timeout[0], $this->timeout[1]))===false)
   {
    if($this->reread["enabled"])
     $this->setTimeout($timeout); //restore original timeout

    if($this->error()) //signal interrupt can trigger a failure, errno is 0 returned
     $this->e->add("Connection read failed: unable to select on socket", "error", "[".$this->errno."] ".$this->error);

    return false;
   }

   if($this->reread["enabled"])
   {
    if($this->reread["counter"])
     $this->setTimeout($timeout); //restore original timeout

    $this->reread["counter"]++;
   }

   if($sockets===0) //Timeout: nothing to read
    return "";

   //start reading
   $received="";
   while($data=@socket_read($this->socket, $this->readBytes, $this->readType)) //false fail, "" done reading
   {
    $received.=$data;
    $bytes=strlen($data);
    $this->bytesReceived+=$bytes;

    if($bytes<$this->readBytes) //less then read bytes, done reading
     break;
   }

   if($data===false&&$this->error()) //signal interrupt can trigger a failure, errno is 0 returned, in this case the received buffer will still be returned
   {
    $this->e->add("Connection read failed from: \"".($this->host?$this->host.":".$this->port:$this->file)."\"", "error", "[".$this->errno."] ".$this->error);
    return false;
   }

   return $received;
  }

  function write($data)
  {
   if(!$this->socket)
   {
    $this->e->add("Socket not active, refusing to write", "error");
    return false;
   }

   if($this->reread["enabled"])
    $this->resetReRead();

   $length=strlen($data);

   $read=NULL;
   $write=Array($this->socket);
   $exception=NULL;

   if(($sockets=@socket_select($read, $write, $exception, $this->timeout[0], $this->timeout[1]))===false)
   {
    if($this->error()) //signal interrupt can trigger a failure, errno is 0 returned
     $this->e->add("Connection write failed: unable to select on socket", "error", "[".$this->errno."] ".$this->error);

    return false;
   }

   if($sockets===0) //Timeout: write failed
    return false;

   if(($bytes=@socket_write($this->socket, $data, $length))===false)
   {
    if($this->error()) //signal interrupt can trigger a failure, errno is 0 returned
     $this->e->add("Connection write failed to: \"".($this->host?$this->host.":".$this->port:$this->file)."\"", "error", "[".$this->errno."] ".$this->error);

    return false;
   }

   $this->bytesSend+=$bytes;
   return true;
  }



  function close($message=false)
  {
   if(!$this->socket&&!$this->socket6)
   {
    $this->e->add("Socket not active, refusing to close", "error");
    return false;
   }

   if($message)
    $this->write($message);

   if($this->reread["enabled"])
    $this->resetReRead();

   if($this->socket)
   {
    @socket_shutdown($this->socket);
    @socket_close($this->socket);
    $this->socket=false;
   }

   if($this->socket6)
   {
    @socket_shutdown($this->socket6);
    @socket_close($this->socket6);
    $this->socket6=false;
   }

   if($this->mode=="connection")
    $this->connectionTime=microtime(true)-$this->startTime;
   else if($this->mode=="listen")
   {
    if($this->file&&file_exists($this->file)) //remove listening file socket
    {
     if(!@unlink($this->file))
     {
      $this->e->add("Unable to remove socket file: \"".$this->file."\"", "error");
      return false;
     }
    }
   }

   return true;
  }



  function fromIP($ip)
  {
   $nr=0;
   $subnet=false;
   if(preg_match("/^([0-9\.]+)(\/(\d+)|\.\*)?$/", $ip, $m)) //ipv4 ip/subnet
   {
    $ip=explode(".", $m[1]);

    if(isset($m[3])) //subnet
     $subnet=$m[3];
    else if(isset($m[2])) //wildcard
     $subnet=8*sizeof($ip);

    if(($l=sizeof($ip))<4) //short ip mode (e.g. 0.0.*)
     array_splice($ip, $l, 0, array_fill($l, 4-$l, 0)); //fill with zeros

    for($c=0;$c<4;$c++)
    {
     if($c<3) //get class divisor
      $class=bcpow(2, 8*(3-$c));
     else
      $class=1;

     $nr=bcadd($nr, bcmul($ip[$c], $class)); //add class value
    }

    if($subnet!==false) //subnet ::/0
    {
     $start=$nr;
     $end=bcadd($start, bcsub(bcpow(2, 32-$subnet), 1)); //add subnet end
     $nr=Array($start, $end);
    }
   }
   else if(preg_match("/^([0-F:]+)(\/(\d+)|:\*)?$/i", $ip, $m)) //ipv6 ip/subnet
   {
    $ip=explode(":", $m[1]);

    if(($l=sizeof($ip))<8) //short ip mode (e.g. ::1)
    {
     for($c=1;$c<$l;$c++) //find shortcut
     {
      if($ip[$c]==="")
      {
       array_splice($ip, $c, 1, array_fill($c, 9-$l, 0)); //fill with zeros
       break;
      }
     }
    }

    if(isset($m[3])) //subnet
     $subnet=$m[3];
    else if(isset($m[2])) //wildcard
    {
     $subnet=16*sizeof($ip);

     if($subnet==128) //::* fix
      $subnet=112;
    }

    if(($l=sizeof($ip))<8) //short ip mode (e.g. 0:0:*)
     array_splice($ip, $l, 0, array_fill($l, 8-$l, 0)); //fill with zeros

    for($c=0;$c<8;$c++)
    {
     if($c<7) //get class divisor
      $class=bcpow(2, 16*(7-$c));
     else
      $class=1;

     $nr=bcadd($nr, bcmul(hexdec($ip[$c]), $class)); //add class value
    }

    if($subnet!==false) //subnet ::/0
    {
     $start=$nr;
     $end=bcadd($start, bcsub(bcpow(2, 128-$subnet), 1)); //add subnet end
     $nr=Array($start, $end);
    }
   }

   return $nr;
  }

  function toIP($nr, $type=false, $short=true)
  {
   $ip=Array();
   if($type==4||($type==false&&$nr<=4294967296)) //forced ipv4 or autodetected
   {
    for($c=0;$c<4;$c++)
    {
     if($c<3) //get class divisor
      $class=bcpow(2, 8*(3-$c));
     else
      $class=1;

     $div=bcdiv($nr, $class, 0); //get class value
     $ip[]=$div;

     if($div) //remove class value for next iteration
      $nr=bcmod($nr, $class);
    }

    $ip=implode(".", $ip); //make proper ip
   }
   else //forced ipv6 or autodetected
   {
    for($c=0;$c<8;$c++)
    {
     if($c<7) //get class divisor
      $class=bcpow(2, 16*(7-$c));
     else
      $class=1;

     $div=bcdiv($nr, $class, 0); //get class value
     $ip[]=dechex($div);

     if($div) //remove class value for next iteration
      $nr=bcmod($nr, $class);
    }

    $ip=implode(":", $ip); //make proper ip

    if($short) //short mode, convert subsequent 0:0:0 to ::
    {
     $ip=preg_replace("/(^|:)(0:)+/", "\\1:", $ip, 1);

     if(substr($ip, -3)=="::0") //ends with :0
      $ip=substr($ip, 0, -1); //remove trailing zero
     else if($ip[0]==":") //begins with 0:
      $ip=":".$ip; //add double colon prefix
    }
   }

   return $ip;
  }

  function inSubnet($subnet, $ip)
  {
   $subnet=Socket::fromIP($subnet); //get subnet ip range

   if(!is_array($subnet)) //must be a proper array
    return false; //can't validate subnet, not in subnet

   $ip=Socket::fromIP($ip); //get ip

   if($subnet[0]<=$ip&&$ip<=$subnet[1]) //ip within subnet
    return true;

   return false;
  }


  function __destruct()
  {
   if($this->socket||$this->socket6)
    $this->close();
  }

 }

?>
